import collections
import pathlib

import imageio
import numpy as np
from PIL import Image


DiscreteSpace = collections.namedtuple('DiscreteSpace', 'n')
BoxSpace = collections.namedtuple('BoxSpace', 'low, high, shape, dtype')
DictSpace = collections.namedtuple('DictSpace', 'spaces')


class AttrDict(dict):

  __getattr__ = dict.__getitem__


class staticproperty:

  def __init__(self, function):
    self.function = function

  def __get__(self, instance, owner=None):
    return self.function()


class World:

  def __init__(self, area, seed=None):
    self._terrain = np.zeros(area, np.uint8)
    self._material_names = {0: None}
    self._material_ids = {None: 0}
    self._objects = [None]
    self._coords = np.zeros(area, np.uint32)
    self._random = np.random.RandomState(seed)

  @property
  def area(self):
    return self._terrain.shape

  @property
  def random(self):
    return self._random

  @property
  def objects(self):
    yield from (obj for obj in self._objects if obj)

  def reset(self, seed=None):
    # TODO: Not really needed. Can just create new instance.
    self._random = np.random.RandomState(seed)
    self._terrain[:] = 0
    self._objects = [None]
    self._coords[:] = 0

  def __setitem__(self, pos, material):
    if material not in self._material_ids:
      id_ = len(self._material_ids)
      self._material_ids[material] = id_
      self._material_names[id_] = material
    self._terrain[pos] = self._material_ids[material]

  def __getitem__(self, pos):
    if _inside((0, 0), pos, self.area):
      material = self._material_names[self._terrain[tuple(pos)]]
    else:
      material = None
    if not (0 <= pos[0] < self._coords.shape[0]):
      obj = False
    elif not (0 <= pos[1] < self._coords.shape[1]):
      obj = False
    else:
      obj = self._objects[self._coords[tuple(pos)]]
    return material, obj

  def nearby(self, pos, distance):
    # TODO: Return both nearby materials and objects.
    ids = set(self._terrain[
        pos[0] - distance: pos[0] + distance,
        pos[1] - distance: pos[1] + distance].flatten().tolist())
    return tuple(self._material_names[x] for x in ids)

  def count(self, material):
    if material not in self._material_ids:
      return 0
    return (self._terrain == self._material_ids[material]).sum()

  def add(self, obj):
    assert hasattr(obj, 'pos')
    obj.pos = np.array(obj.pos)
    assert self[obj.pos][1] is None
    self._coords[tuple(obj.pos)] = len(self._objects)
    self._objects.append(obj)

  def remove(self, obj):
    self._objects[self._coords[tuple(obj.pos)]] = None
    self._coords[tuple(obj.pos)] = 0

  def move(self, obj, pos):
    pos = np.array(pos)
    assert self[pos][1] is None
    self._coords[tuple(pos)] = self._coords[tuple(obj.pos)]
    self._coords[tuple(obj.pos)] = 0
    obj.pos = pos


class Textures:

  def __init__(self, directory):
    self._originals = {}
    self._textures = {}
    for filename in pathlib.Path(directory).glob('*.png'):
      image = imageio.imread(filename.read_bytes())
      image = image.transpose((1, 0) + tuple(range(2, len(image.shape))))
      self._originals[filename.stem] = image
      self._textures[(filename.stem, image.shape[:2])] = image

  def get(self, name, size):
    size = int(size[0]), int(size[1])
    key = name, size
    if key not in self._textures:
      image = self._originals[name]
      image = Image.fromarray(image)
      image = image.resize(size[::-1], resample=Image.NEAREST)
      image = np.array(image)
      self._textures[key] = image
    return self._textures[key]


class GlobalView:

  pass


class UncoverView:

  pass


class LocalView:

  def __init__(self, world, textures, unit, grid):
    self._world = world
    self._textures = textures
    self._unit = np.array(unit)
    self._grid = np.array(grid)
    self._offset = self._grid // 2
    self._area = np.array(self._world.area)
    self._center = None

  def __call__(self, player):
    self._center = np.array(player.pos)
    canvas = np.zeros(tuple(self._grid * self._unit) + (3,), np.uint8) + 127
    for x in range(self._grid[0]):
      for y in range(self._grid[1]):
        pos = self._center + np.array([x, y]) - self._offset
        if not _inside((0, 0), pos, self._area):
          continue
        texture = self._textures.get(self._world[pos][0], self._unit)
        _draw(canvas, np.array([x, y]) * self._unit, texture)
    for obj in self._world.objects:
      pos = obj.pos - self._center + self._offset
      if not _inside((0, 0), pos, self._grid):
        continue
      texture = self._textures.get(obj.texture, self._unit)
      _draw_alpha(canvas, pos * self._unit, texture)
    return canvas


class ItemView:

  def __init__(self, textures, unit, grid):
    self._textures = textures
    self._unit = np.array(unit)
    self._grid = np.array(grid)

  def __call__(self, inventory):
    canvas = np.zeros(tuple(self._grid * self._unit) + (3,), np.uint8)
    for index, (item, amount) in enumerate(inventory.items()):
      if amount < 1:
        continue
      self._item(canvas, index, item)
      self._amount(canvas, index, amount)
    return canvas

  def _item(self, canvas, index, item):
    pos = index % self._grid[0], index // self._grid[0]
    pos = (pos * self._unit + 0.1 * self._unit).astype(np.int32)
    texture = self._textures.get(item, 0.8 * self._unit)
    _draw_alpha(canvas, pos, texture)

  def _amount(self, canvas, index, amount):
    pos = index % self._grid[0], index // self._grid[0]
    pos = (pos * self._unit + 0.4 * self._unit).astype(np.int32)
    if amount == float('inf'):
      text = 'infinity'
    elif amount in (1, 2, 3, 4, 5):
      text = str(amount)
    else:
      text = 'unknown'
    texture = self._textures.get(text, 0.6 * self._unit)
    _draw_alpha(canvas, pos, texture)


def _inside(lhs, mid, rhs):
  return (lhs[0] <= mid[0] < rhs[0]) and (lhs[1] <= mid[1] < rhs[1])

def _draw(canvas, pos, texture):
  (x, y), (w, h) = pos, texture.shape[:2]
  if texture.shape[-1] == 4:
    texture = texture[..., :3]
  canvas[x: x + w, y: y + h] = texture

def _draw_alpha(canvas, pos, texture):
  (x, y), (w, h) = pos, texture.shape[:2]
  if texture.shape[-1] == 4:
    alpha = texture[..., 3:].astype(np.float32) / 255
    texture = texture[..., :3].astype(np.float32) / 255
    current = canvas[x: x + w, y: y + h].astype(np.float32) / 255
    blended = alpha * texture + (1 - alpha) * current
    texture = (255 * blended).astype(np.uint8)
  canvas[x: x + w, y: y + h] = texture



import numpy as np

from . import objects


MAP = """
SSSSSSSSSSSSSS
SWWWW   T    S
SWW          S
S         T  S
S         T  S
S            S
S            S
S   T        S
S            S
S            S
S         T  S
S   WW       S
S  WWWWWWW   S
SSSSSSSSSSSSSS
"""


def generate_world(world, player, num_cows):
  lines = MAP.split('\n')
  lines = [line.strip() for line in lines if line]
  for y, line in enumerate(lines):
    for x, char in enumerate(line):
      world[x, y] = {
          ' ': 'grass',
          'T': 'tree',
          'S': 'stone',
          'W': 'water',
      }[char]
  empties = []
  for x in range(world.area[0]):
    for y in range(world.area[1]):
      dist = np.abs(player.pos - np.array([x, y])).sum()
      if world[x, y][0] == 'grass' and dist > 5:
        empties.append((x, y))
  for _ in range(num_cows):
    index = world.random.randint(0, len(empties))
    pos = empties[index]
    del empties[index]
    world.add(objects.Cow(world, pos, player))



import argparse

import imageio
try:
  import pygame
except ImportError:
  print('Please install the pygame package to use the GUI.')
  raise

import cowherd


def main():
  parser = argparse.ArgumentParser()
  parser.add_argument('--seed', type=int, default=None)
  parser.add_argument('--view', type=int, nargs=2, default=(7, 7))
  parser.add_argument('--num_cows', type=int, default=3)
  parser.add_argument('--length', type=int, default=None)
  parser.add_argument('--health', type=int, default=5)
  parser.add_argument('--window', type=int, nargs=2, default=(800, 800))
  parser.add_argument('--record', type=str, default=None)
  parser.add_argument('--fps', type=int, default=5)
  args = parser.parse_args()

  keymap = {
      pygame.K_a: 'move_left',
      pygame.K_d: 'move_right',
      pygame.K_w: 'move_up',
      pygame.K_s: 'move_down',
      pygame.K_SPACE: 'do',
      pygame.K_1: 'place_fence',
  }
  print('Actions:')
  for key, action in keymap.items():
    print(f'  {pygame.key.name(key)}: {action}')

  env = cowherd.Env(
      args.view, args.window, args.length, args.num_cows, args.seed)
  env.reset()
  achievements = set()
  duration = 0
  return_ = 0
  was_done = False
  if args.record:
    frames = []
  print('Diamonds exist:', env._world.count('diamond'))

  pygame.init()
  screen = pygame.display.set_mode(args.window)
  clock = pygame.time.Clock()
  running = True
  while running:

    action = None
    pygame.event.pump()
    for event in pygame.event.get():
      if event.type == pygame.QUIT:
        running = False
      elif event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
        running = False
      elif event.type == pygame.KEYDOWN and event.key in keymap.keys():
        action = keymap[event.key]
    if action is None:
      pressed = pygame.key.get_pressed()
      for key, action in keymap.items():
        if pressed[key]:
          break
      else:
        action = 'noop'

    messages = []
    obs, reward, done, _ = env.step(env.action_names.index(action))

    unlocked = {
        name for name, count in env._player.achievements.items()
        if count > 0 and name not in achievements}
    for name in unlocked:
      achievements |= unlocked
      total = len(env._player.achievements.keys())
      messages.append(f'Achievement ({len(achievements)}/{total}): {name}')
    if env._step > 0 and env._step % 100 == 0:
      messages.append(f'Time step: {env._step}')
    if reward:
      messages.append(f'Reward: {reward}')
      return_ += reward
    if done and not was_done:
      was_done = True
      messages.append('Episode done!')
    if messages:
      print('\n'.join(messages), sep='')

    duration += 1
    if args.record:
      frames.append(obs)
    surface = pygame.surfarray.make_surface(
        obs.transpose((1, 0, 2)))

    screen.blit(surface, (0, 0))
    pygame.display.flip()
    clock.tick(args.fps)

  pygame.quit()
  print('Duration:', duration)
  print('Return:', return_)
  if args.record:
    imageio.mimsave(args.record, frames)
    print('Saved', args.record)


if __name__ == '__main__':
  main()



import numpy as np

from . import constants
from . import engine
from . import objects
from . import worldgen


class Env:

  def __init__(
      self, view=(7, 7), size=(64, 64), length=1000, num_cows=3, seed=None):
    view = np.array(view if hasattr(view, '__len__') else (view, view))
    size = np.array(size if hasattr(size, '__len__') else (size, size))
    unit = size // view
    self._size = size
    self._length = length
    self._num_cows = num_cows
    self._seed = seed
    self._episode = 0
    self._world = engine.World((14, 14))
    self._textures = engine.Textures(constants.root / 'assets')
    item_rows = int(np.ceil(len(constants.items) / view[0]))
    self._local_view = engine.LocalView(
        self._world, self._textures, unit,
        [view[0], view[1] - item_rows])
    self._item_view = engine.ItemView(
        self._textures, unit, [view[0], item_rows])
    self._border = (size - unit * view) // 2
    self._step = None
    self._player = None
    self._milked = None

  @property
  def observation_space(self):
    return engine.BoxSpace(0, 255, tuple(self._size) + (3,), np.uint8)

  @property
  def action_space(self):
    return engine.DiscreteSpace(len(constants.actions))

  @property
  def action_names(self):
    return constants.actions

  def reset(self):
    center = (self._world.area[0] // 2, self._world.area[1] // 2)
    self._step = 0
    self._episode += 1
    self._world.reset(seed=hash((self._seed, self._episode)) % 2 ** 32)
    self._player = objects.Player(self._world, center)
    self._player.inventory['fence'] = float('inf')
    self._world.add(self._player)
    self._milked = 0
    worldgen.generate_world(self._world, self._player, self._num_cows)
    return self._obs()

  def step(self, action):
    self._step += 1
    # Copy object list so new added objects are not updated right away.
    for obj in list(self._world.objects):
      if obj is self._player:
        obj.update(action)
      else:
        obj.update()
    obs = self._obs()

    if self._milked < self._player.achievements['milk_cow']:
      self._milked = self._player.achievements['milk_cow']
      reward = 1.0
    else:
      reward = 0.0
    done = self._length and self._step >= self._length
    info = {
        'inventory': self._player.inventory.copy(),
        'achievements': self._player.achievements.copy(),
        'discount': 1.0,
    }
    return obs, reward, done, info

  def render(self):
    canvas = np.zeros(tuple(self._size) + (3,), np.uint8)
    local_view = self._local_view(self._player)
    item_view = self._item_view(self._player.inventory)
    view = local_view
    view = np.concatenate([local_view, item_view], 1)
    (x, y), (w, h) = self._border, view.shape[:2]
    canvas[x: x + w, y: y + h] = view
    return canvas.transpose((1, 0, 2))

  def _obs(self):
    return self.render()



import numpy as np

from . import constants


class Object:

  def __init__(self, world, pos):
    self.world = world
    self.pos = np.array(pos)
    self.random = world.random
    self.health = 0

  @property
  def texture(self):
    raise 'unknown'

  @property
  def walkable(self):
    return constants.walkable

  def move(self, direction):
    direction = np.array(direction)
    target = self.pos + direction
    if self.is_free(target):
      self.world.move(self, target)
      return True
    return False

  def is_free(self, target, materials=None):
    materials = self.walkable if materials is None else materials
    material, obj = self.world[target]
    return obj is None and material in materials

  def distance(self, target):
    if hasattr(target, 'pos'):
      target = target.pos
    return np.abs(target - self.pos).sum()

  def toward(self, target, long_axis=True):
    if hasattr(target, 'pos'):
      target = target.pos
    offset = target - self.pos
    dists = np.abs(offset)
    if (dists[0] > dists[1] if long_axis else dists[0] <= dists[1]):
      return np.array((np.sign(offset[0]), 0))
    else:
      return np.array((0, np.sign(offset[1])))

  def random_dir(self):
    dirs = ((-1, 0), (+1, 0), (0, -1), (0, +1))
    return dirs[self.random.randint(0, len(dirs))]


class Player(Object):

  def __init__(self, world, pos):
    super().__init__(world, pos)
    self.facing = (0, 1)
    self.inventory = {item: 0 for item in constants.items}
    self.achievements = {name: 0 for name in constants.achievements}

  @property
  def texture(self):
    return {
        (-1, 0): 'player-left',
        (+1, 0): 'player-right',
        (0, -1): 'player-up',
        (0, +1): 'player-down',
    }[tuple(self.facing)]

  @property
  def walkable(self):
    return constants.walkable + ['lava']

  def update(self, action):
    target = (self.pos[0] + self.facing[0], self.pos[1] + self.facing[1])
    material, obj = self.world[target]
    action = constants.actions[action]
    if action == 'noop':
      pass
    elif action.startswith('move_'):
      self._move(action[len('move_'):])
    elif action == 'do' and obj:
      self._interact(obj)
    elif action == 'do':
      self._collect(target, material)
    elif action.startswith('place_'):
      self._place(action[len('place_'):], target, material)

  def _move(self, direction):
    dirs = dict(left=(-1, 0), right=(+1, 0), up=(0, -1), down=(0, +1))
    self.facing = dirs[direction]
    self.move(self.facing)
    if self.world[self.pos][0] == 'lava':
      self.health = 0

  def _interact(self, obj):
    if isinstance(obj, Cow) and not obj.molken:
      self.inventory['milk'] += 1
      self.achievements['milk_cow'] += 1
      obj.molken = True

  def _collect(self, target, material):
    info = constants.collect.get(material)
    if not info:
      return
    for name, amount in info['require'].items():
      if self.inventory[name] < amount:
        return
    self.world[target] = info['leaves']
    for name, amount in info['receive'].items():
      self.inventory[name] += 1
    self.achievements[f'collect_{material}'] += 1

  def _place(self, name, target, material):
    info = constants.place[name]
    if material not in info['where']:
      return
    if any(self.inventory[k] < v for k, v in info['uses'].items()):
      return
    for item, amount in info['uses'].items():
      self.inventory[item] -= amount
    self.world[target] = name
    self.achievements[f'place_{name}'] += 1

  def _make(self, name):
    nearby = self.world.nearby(self.pos, 2)
    info = constants.make[name]
    if not all(util in nearby for util in info['nearby']):
      return
    if any(self.inventory[k] < v for k, v in info['uses'].items()):
      return
    for item, amount in info['uses'].items():
      self.inventory[item] -= amount
    self.inventory[name] += 1
    self.achievements[f'make_{name}'] += 1


class Cow(Object):

  def __init__(self, world, pos, player):
    super().__init__(world, pos)
    self.health = 3
    self.player = player
    self.molken = False

  @property
  def texture(self):
    if self.molken:
      return 'cow-molken'
    else:
      return 'cow'

  def update(self):
    near = self.distance(self.player) < 3
    if near and self.random.uniform() < 0.8:
      away = -self.toward(self.player, self.random.uniform() < 0.6)
      if self.is_free(self.pos + away):
        self.move(away)
        return
      dirs = [(-1, 0), (+1, 0), (0, -1), (0, +1)]
      dirs = [np.array(x) for x in dirs]
      dirs = [x for x in dirs if self.is_free(self.pos + x)]
      dirs = [x for x in dirs if self.player.distance(self.pos + x) > 1]
      dirs = dirs if dirs else [self.random_dir()]
      self.move(dirs[self.random.randint(0, len(dirs))])
    elif near or self.random.uniform() < 0.3:
      self.move(self.random_dir())
